using System;
using System.IO;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using MeshVR;
using MVR.FileManagementSecure;
using AssetBundles;
using SimpleJSON;
using System.Text.RegularExpressions;

namespace Stopper {
	public class AlternativeFuta : MVRScript {
	
	string packageId;
	string packagePath;
	
	// Character skin prefabs
	GameObject femaleSkinsPrefab;
	GameObject maleSkinsPrefab;
	
	GameObject femaleSkinsContainer;
	
	// Stuff for the new 'DAZCharacter'
	// that has our new modified skin
	GameObject newCharacterContainer;
	DAZCharacter newCharacter;
	
	DAZCharacterSelector characterSelector;
	
	DAZCharacterRun characterRun;
	
	DAZMergedSkinV2 skin;
		
	DAZMergedMesh mergedMesh;
	
	DAZMesh genMesh;
	
	GameObject femaleSkinBase;
	
	DAZBone gen1Bone;
	DAZBone gen2Bone;
	DAZBone gen3Bone;
	
	string futaSkinName = "AltFutaSkin";
	
	// Previous
	DAZCharacter previousSelected;
	
	
	Dictionary<string, string> femaleSkinsWithMatchingFuta = new Dictionary<string, string>()
	{
		{ "renkayla", "FutaKayla" },
		{ "renlexi", "FutaLexi" },
		{ "renmaria", "FutaMaria" },
		{ "renmia", "FutaMia" },
		{ "rensimone", "FutaSimone" },
	};
	
	// We also instantiate a male character to steal
	// some data from during initialization
	GameObject maleCharacterContainer;
	GameObject maleSkinBase;
	DAZMergedSkinV2 maleSkin;
	DAZMergedMesh maleMergedMesh;
	
	List<GameObject> maleObjects = new List<GameObject>();
	
	// UI Stuff
	GameObject UIContainer;
	
	GameObject materialUIPrefab;
	MVRScriptUI scriptUI;
	
	JSONStorableBool hideFemaleGens;
	JSONStorableBool useTransparentShader;
	JSONStorableBool drawMalePelvis;
	
	bool skinLoaded = false;
	bool useVaginaAsPelvisBase = false;
	bool overrideFemalePelvisTextures = false;
	
	// Mat Options
	
	MaterialOptionsWrapper genMaterialOptions;
	Transform genMaterialOptionsUI;
	
	MaterialOptionsWrapper pelvisMaterialOptions;
	Transform pelvisMaterialOptionsUI;
	
	MaterialOptionsWrapper outerPelvisMaterialOptions;
	Transform outerPelvisMaterialOptionsUI;
	
	// Morph Stuff
	//float genBoneYOffsetDefault = 0.028f;
	//float genBoneZOffsetDefault = -0.016f
	
	float genBoneYOffsetDefault = 0.023f;
	float genBoneZOffsetDefault = -0.013f;
	
	DAZMorph genYOffsetMorph;
	DAZMorph genZOffsetMorph;
	DAZMorph genYMeshOffsetMorph;
	DAZMorph genZMeshOffsetMorph;
	
	JSONStorableFloat genYOffetValue;
	JSONStorableFloat genZOffetValue;
	JSONStorableFloat genYMeshOffetValue;
	JSONStorableFloat genZMeshOffetValue;
	
	// Saving and loading
	JSONClass savedSceneConfig;
	
	JSONClass[] oldSkinOpts;
	JSONClass oldTextureControlJson;
	
	// Physics stuff
	GameObject maleGen1;
	
	List<string> genAutoColliderIgnoreNames = new List<string>{
			// Front pelvis
			"AutoColliderpelvisF1","AutoColliderpelvisF2","AutoColliderpelvisF3",
			"AutoColliderpelvisF4","AutoColliderpelvisF5","AutoColliderpelvisF6",
			// Inner pelvis
			"AutoColliderpelvisFL1","AutoColliderpelvisFL2","AutoColliderpelvisFL3","AutoColliderpelvisFL4",
			"AutoColliderpelvisFR1","AutoColliderpelvisFR2","AutoColliderpelvisFR3","AutoColliderpelvisFR4"
	};
	List<AutoCollider> genAutoColliderIgnoreList = new List<AutoCollider>();
	
	Dictionary<AutoCollider, AutoColliderConfig> maleAutoColliderConfigs = new Dictionary<AutoCollider, AutoColliderConfig>();
	
	public class AutoColliderConfig {
		public int targetVertex = -1;
		public int anchorVertex1 = -1;
		public int anchorVertex2 = -1;
		public int oppositeVertex = -1;
	}
	
	// Small wrappers that help us get access to some protected members
	public class DAZSkinV2Wrapper : DAZSkinV2 {
		public void Copy(DAZSkinV2 other){
			this.sceneGeometryId = other.sceneGeometryId;
			this._numBones = other.numBones;
			this.nodes = other.nodes;
		}
	}
	
	// This is used to detect when the selected character is changed
	public class DisableListener : MonoBehaviour {
		public AlternativeFuta script;
		
		public void OnDisable(){
			if(script == null) return;
			
			
		
			// Restore modified auto colliders
			if(script.maleAutoColliderConfigs != null){
				foreach(var entry in script.maleAutoColliderConfigs){
					AutoCollider ac = entry.Key;
					AutoColliderConfig config = entry.Value;
					
					ac.targetVertex = config.targetVertex;
					ac.anchorVertex1 = config.anchorVertex1;
					ac.anchorVertex2 = config.anchorVertex2;
					ac.oppositeVertex = config.oppositeVertex;
					
					ac.AutoColliderSizeSet(true);
				}
			}
			
			script.SetMaleTransformsActive(false);
		}
	}

	public class MaterialOptionsWrapper : DAZCharacterMaterialOptions {
		public void SetParams(){
			this.SetAllParameters();
		}
		
		public void CopyMaterialParams(MaterialOptions src){
			color1CurrentColor = src.color1CurrentColor;
			color2CurrentColor = src.color2CurrentColor;
			color3CurrentColor = src.color3CurrentColor;
			
			param1CurrentValue = src.param1CurrentValue;
			param2CurrentValue = src.param2CurrentValue;
			param3CurrentValue = src.param3CurrentValue;
			param4CurrentValue = src.param4CurrentValue;
			param5CurrentValue = src.param5CurrentValue;
			param6CurrentValue = src.param6CurrentValue;
			param9CurrentValue = src.param9CurrentValue;
			param10CurrentValue = src.param10CurrentValue;
			
			SetParams();
		}
		
		public void ClearCustomTextures(){
			if(customTexture1UrlJSON != null) customTexture1UrlJSON.Clear();
			if(customTexture2UrlJSON != null) customTexture2UrlJSON.Clear();
			if(customTexture3UrlJSON != null) customTexture3UrlJSON.Clear();
			if(customTexture4UrlJSON != null) customTexture4UrlJSON.Clear();
			if(customTexture5UrlJSON != null) customTexture5UrlJSON.Clear();
			if(customTexture6UrlJSON != null) customTexture6UrlJSON.Clear();
		}
	}
	
	public override void Init() {
		packageId = GetPackageId(this);
		packagePath = string.IsNullOrEmpty(packageId) ? "" : $"{packageId}:/";
	
		// Initialize some references. 		
				
		characterRun = containingAtom.GetComponentInChildren<DAZCharacterRun>();
		
		gen1Bone = characterRun.bones.GetDAZBone("Gen1");
		gen2Bone = characterRun.bones.GetDAZBone("Gen2");
		gen3Bone = characterRun.bones.GetDAZBone("Gen3");
		
		GetMaleTransforms();
		SaveAutoColliderConfigs();
		
		characterSelector = containingAtom.GetComponentInChildren<DAZCharacterSelector>();
		previousSelected = characterSelector.selectedCharacter;
		
		// TODO: Replace these with hard-coded paths
		foreach (GameObject go in Resources.FindObjectsOfTypeAll(typeof(GameObject)) as GameObject[]){
			if(go.name =="FemaleCharactersPrefab"){
				femaleSkinsPrefab = go;
			}
			
			if(go.name =="MaleCharactersPrefab"){
				maleSkinsPrefab = go;
			}
		}
		
		foreach(GameObject go in GetAllChildren(containingAtom.gameObject)){
			if(go.name.Contains("FemaleCharactersPrefab")){
				femaleSkinsContainer = go;
			}
		}
		
		// Store previous material options
		try {		
			if(previousSelected != null){
				DAZCharacterMaterialOptions[] oldOptions = previousSelected.GetComponentsInChildren<DAZCharacterMaterialOptions>();
				oldSkinOpts = new JSONClass[oldOptions.Length];
				for(int j = 0; j < oldOptions.Length; j++){
					var json = oldOptions[j].GetJSON(true,true,true);
					oldSkinOpts[j] = json;
				}
			}
		} catch (Exception e) {
			SuperController.LogError($"Error storing previous skin options. {e}");
		}
		
		// Store other material settings in DAZCharacterTextureControl
		var oldTextureControl = previousSelected.GetComponentInChildren<DAZCharacterTextureControl>();
		if(oldTextureControl != null){
			oldTextureControlJson = oldTextureControl.GetJSON(true,true,true);
		}
		
		
		// Select the base skin.
		// We'll try to use whatever selected, otherwise we choose 'rensimone' arbitrarily
		string baseSkinName = null;
		if(characterSelector.selectedCharacter != null) {
			baseSkinName = characterSelector.selectedCharacter.name;
			
		}
		
		if(baseSkinName == null || baseSkinName == futaSkinName){
			baseSkinName = "rensimone";
		}
		
		foreach(GameObject go in GetAllChildren(femaleSkinsPrefab)){
			if(go.name == baseSkinName){
				femaleSkinBase = go;
			}
		}
		
		if(femaleSkinBase == null){
			SuperController.LogMessage($"Unable to find female skin preset for '{baseSkinName}'");
			return;
		}
		
		
		string maleSkinName = null;
		femaleSkinsWithMatchingFuta.TryGetValue(baseSkinName, out maleSkinName);
		if(maleSkinName == null){
			maleSkinName = "FutaSimone";
		}
		
		foreach(GameObject go in GetAllChildren(maleSkinsPrefab)){
			if(go.name == maleSkinName){
				maleSkinBase = go;
			}
		}
		
		foreach(AutoCollider ac in containingAtom.GetComponentsInChildren<AutoCollider>()){
			if(genAutoColliderIgnoreNames.Contains(ac.name)){
				genAutoColliderIgnoreList.Add(ac);
			}
		}
		
		InitMainUI();
	}
	
	
	private void GetMaleTransforms(){
		// TODO: Use hardcoded paths for these
		foreach(GameObject go in GetAllChildren(containingAtom.gameObject)){
			if((go.name.Contains("penis") || go.name.Contains("testes")) 
				&& go.name.Contains("Control")){
				maleObjects.Add(go);
			}
			if(	go.name == "Gen1" ){
				maleGen1 = go;
				maleObjects.Add(go);
			} else if( go.name == "Gen2"
					|| go.name == "Gen3"
					|| go.name == "Testes"){
				maleObjects.Add(go);
			}
		}
	}
	
	private void SaveAutoColliderConfigs(){
		foreach(AutoCollider ac in maleGen1.GetComponentsInChildren<AutoCollider>(true)){
			maleAutoColliderConfigs[ac] = new AutoColliderConfig{ 
				targetVertex = ac.targetVertex,
				anchorVertex1 = ac.anchorVertex1,
				anchorVertex2 = ac.anchorVertex2,
				oppositeVertex = ac.oppositeVertex,
			};
		}
	}
	
	public void Start() {
		LoadSkin();
	}
	
	private void LoadSkin(){
		StartCoroutine("AddMaleGenitalia");
		StartCoroutine("InitMaterialUI");
	}
	
	private void ReloadSkin(){
		if(skinLoaded){
			skinLoaded = false;
			savedSceneConfig = GetJSON(true,true,true);
			UnloadSkin();
			LoadSkin();
		}
	}
	
	private IEnumerator AddMaleGenitalia(){
		
		
		/* 	The VAM skin hierarchy includes a DAZMergedSkinV2, which uses a DAZMergedMesh. 
			'Merged' here refers to the fact that genitals have their own DAZMesh that is 
			grafted onto the main mesh. There is code for 2 graft meshes and, luckily, the 
			female model only uses one. 
		
			The first thing we'll do is set up a new skin. This requires a few high-level steps:
			1. Create a new DAZCharacter off of a prefab
				- This will create most of our DAZMergedMesh and our DAZMergedSkinV2
			2. Add a DAZMesh and DAZSkinV2 for the penis prior to the main skin initializing
				- The DAZMesh we get from a file
				- The DAZSkinV2 we'll copy from a male skin
			3. Merge the DAZMergedMesh and our DAZMergedSkinV2
			4. Set the materials
		*/
		
		// Initialize the male skin that we'll steal some data from.
		// Specifically, we want the node weights, which determine how 
		// each vertex is weiged to the DAZBones of the skeleton
		maleCharacterContainer = UnityEngine.Object.Instantiate<GameObject>(maleSkinBase);
		
		// Enable the object. This will cause the DAZCharacter to load skin
		// artifacts from assetBundles
		maleCharacterContainer.SetActive(true);
		
		DAZMergedSkinV2 maleMergedSkin = null;
		while(maleMergedSkin == null){
			maleMergedSkin = maleCharacterContainer.GetComponentInChildren<DAZMergedSkinV2>();
			yield return null;
		}
		maleCharacterContainer.SetActive(false);
		maleMergedMesh = (maleMergedSkin.dazMesh as DAZMergedMesh);
			
	
		// Initialize the new skin, which will initially include just the DAZCharacter
		newCharacterContainer = UnityEngine.Object.Instantiate<GameObject>(femaleSkinBase);
		newCharacterContainer.SetActive(false);
		newCharacterContainer.name = futaSkinName;
		newCharacterContainer.transform.parent = femaleSkinsContainer.transform;
		
		var disableListener = newCharacterContainer.AddComponent<DisableListener>();
		disableListener.script = this;
		
		newCharacter = newCharacterContainer.GetComponentInChildren<DAZCharacter>();
		
		newCharacter.containingAtom = containingAtom;
		newCharacter.rootBonesForSkinning = characterSelector.rootBones;
		
		
		// Once a DAZSkinV2 is initialized, all the vertices are copied into ComputeBuffers
		// and I don't know how to reset those. This means that we need to do all of our mesh
		// modification prior to initialization.
		
		// Initialization happens in the onLoadedHandler, so in order to wedge ourselves in 
		// between the DAZCharacter loading the assets and initialization, we'll just also
		// add ourselves to the onLoadedHandler (which DAZCharacter will later combine with its own)
		newCharacter.onLoadedHandlers = () => {
			try {
				// Skin should be loaded (but not initialized) by this point
				skin = newCharacterContainer.GetComponentInChildren<DAZMergedSkinV2>();	
				mergedMesh = (skin.dazMesh as DAZMergedMesh);
				skin.name = futaSkinName;
		
				// Load our mesh from file. This is the only way I found to initialize the
				// vertices manually. 
				// The file I'm using came from using StoreToBinaryFile, which required me to 
				// disable file IO restrictions. (Luckily LoadFromBinaryFile doesn't have those)
				genMesh = mergedMesh.gameObject.AddComponent<DAZMesh>();
				genMesh.LoadFromBinaryFile($"{GetPluginPath(this,containingAtom)}/male_gen_dazmesh");
				
				// The meshgraft includes instructions on which vertices should
				// be 'glued' to the base model. We're not bothering with this for now.
				// TODO: Set this...
				genMesh.meshGraft = maleMergedMesh.graftMesh.meshGraft;
				
				// Don't bother grafting the pelvis to the main body to prevent it deforming
				if(!drawMalePelvis.val){
					genMesh.meshGraft.vertexPairs = new DAZMeshGraftVertexPair[0];
				}
				genMesh.meshGraft.hiddenPolys = new int[0];
				
				genMesh.graftTo = mergedMesh.targetMesh;
				
				genMesh.Init();
			
				// Merge the skin! 
				// This should result in the mergedMesh copying all of our gen mesh's 
				// vertices. 
				mergedMesh.Merge();
				
				
				// DAZSkinV2 also has a merge concept, which copies over the bone weights
				// that determine how the mesh is skinned to the skeleton.
				
				// For this we'll use the male skin we created earlier and copy its verts
				DAZSkinV2 maleGenSkin = null;
				foreach(var skinv2 in maleCharacterContainer.GetComponentsInChildren<DAZSkinV2>()){
					if(skinv2.dazMesh == maleMergedMesh.graftMesh){
						maleGenSkin = skinv2;
					}
				}
				DAZSkinV2Wrapper genSkinv2 = mergedMesh.gameObject.AddComponent<DAZSkinV2Wrapper>();
				genSkinv2.Copy(maleGenSkin);
				
				// Merge the skins.
				skin.Merge();
				
				
				// Adjust the male auto colliders
				foreach(var entry in maleAutoColliderConfigs){
					AutoCollider ac = entry.Key;
					AutoColliderConfig config = entry.Value;
					
					if(ac.targetVertex >=0) ac.targetVertex = config.targetVertex + mergedMesh.numGraftBaseVertices;
					if(ac.anchorVertex1 >=0) ac.anchorVertex1 = config.anchorVertex1 + mergedMesh.numGraftBaseVertices;
					if(ac.anchorVertex2 >=0) ac.anchorVertex2 = config.anchorVertex2 + mergedMesh.numGraftBaseVertices;
					if(ac.oppositeVertex >=0) ac.oppositeVertex = config.oppositeVertex + mergedMesh.numGraftBaseVertices;
				}
			
				// Set the materials
				
				Material[] originalMaterials = skin.GPUmaterials;
				skin.CopyMaterials();
				for (int i = 0; i < skin.dazMesh.materials.Length; i++)	{
					skin.materialsEnabled[i] = true;
					skin.dazMesh.materialsEnabled[i] = true;
				
				if(i < originalMaterials.Length){
						skin.GPUmaterials[i] = originalMaterials[i];
					} else {
						skin.GPUmaterials[i] = new Material(Shader.Find("Custom/Discard"));
					}
				}
				// Female gens
				if(useVaginaAsPelvisBase) {
					skin.GPUmaterials[31] = new Material(skin.GPUmaterials[28]);
				} else {
					skin.GPUmaterials[31] = new Material(skin.GPUmaterials[20]);
				}
				
				// Penis
				skin.GPUmaterials[30] = maleMergedSkin.GPUmaterials[29];
				if(useTransparentShader.val){
					skin.GPUmaterials[30].shader = Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff");
				}
				
				// Pelvis
				if(drawMalePelvis.val){
					skin.GPUmaterials[29]= new Material(maleMergedSkin.GPUmaterials[28]);
					if(useTransparentShader.val){
						skin.GPUmaterials[29].shader = Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff");
					}
				}
				
				skin.FlushBuffers();
				
			} catch (Exception e) {
				SuperController.LogError($"Error initializing merged futa skin. {e}");
			}
		}; // End onLoadedHandlers


		// Set the new character as selected
		try {
		
			SetMaleTransformsActive(false); // In case they're enabled for some reason
			characterSelector.selectedCharacter = newCharacter;
		} catch (Exception e) {
			SuperController.LogError($"Error switching to new skin. {e}");
		}
		
		
		// Wait for it do its thing...
		while(newCharacterContainer == null || newCharacter == null || !newCharacter.ready){
			yield return null;
		}
		
		while(skin == null || !skin.wasInit){
			skin = newCharacterContainer.GetComponentInChildren<DAZMergedSkinV2>();
			yield return null;
		}
		
		
		// There are 3 morph banks for the (up to) 3 grafted meshes.
		// We want our 2nd graft (morphBank3) to use the morph's for 
		// The male's 1st graft (maleMorphBank2)
		characterRun.morphBank3 = characterSelector.maleMorphBank2;
		characterSelector.morphsControlFemaleUI.morphBank3 = characterSelector.maleMorphBank2;
		
		// morph banks will try to apply morphs to connected meshes
		// This might be set if the model was previously male. Need to unset it
		characterRun.morphBank3.connectedMesh = null;
		characterRun.ResetMorphs();	
		characterSelector.maleMorphBank2.gameObject.SetActive(true);
		SetMaleTransformsActive(true);
		
		try {		
			SetUpMorphs();
		} catch (Exception e) {
			SuperController.LogError($"Error setting up positioning morphs. {e}");
		}
		
		// Create the material options
		var defaultsMat = maleMergedSkin.GPUmaterials[29];
		if(useTransparentShader.val)
			defaultsMat.shader = Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff");
		
		Material outerPelvisDefault = null;
		if(useVaginaAsPelvisBase) {
			outerPelvisDefault = skin.GPUmaterials[28];
		} else {
			outerPelvisDefault = skin.GPUmaterials[20];
		}
		
		genMaterialOptions = CreateMaterialOptions(30, defaultsMat);
		pelvisMaterialOptions = CreateMaterialOptions(29, defaultsMat);
		outerPelvisMaterialOptions = CreateMaterialOptions(31, outerPelvisDefault);
		
		RestoreSavedMaterialOptions();
		
		// Re-enable the male portions of the model
		skinLoaded = true;
		
		// There's a race condition here that prevents the autocollider
		// update from taking effecti, but I'm not sure what it is.
		// Waiting for a few frames seems to be enough. 
		yield return new WaitForSecondsRealtime(0.5f);
		UpdteAutoColliders();
		
	}
	
	private void InitMainUI(){
		UIDynamicButton genMaterialOptionsButton = CreateButton("Penis Material Options", true);
		genMaterialOptionsButton.button.onClick.AddListener(() => {
			ShowMaterialOptionsUI(genMaterialOptionsUI);
		});
		
		UIDynamicButton pelvisMaterialOptionsButton = CreateButton("Male Pelvis Material Options", true);
		pelvisMaterialOptionsButton.button.onClick.AddListener(() => {
			ShowMaterialOptionsUI(pelvisMaterialOptionsUI);
		});
		
		UIDynamicButton outerPelvisMaterialOptionsButton = CreateButton("Female Pelvis Material Options", true);
		outerPelvisMaterialOptionsButton.button.onClick.AddListener(() => {
			ShowMaterialOptionsUI(outerPelvisMaterialOptionsUI);
		});
		
		UIDynamicButton reload = CreateButton("Reload Skin");
		reload.button.onClick.AddListener(() => {
			ReloadSkin();
		});
		
		CreateTextField(new JSONStorableString("hideFemaleGensText", "Don't render female gens, and instead renders the base G2F mesh. Does not have an anus."), false).height = 10;
		
		hideFemaleGens = new JSONStorableBool("Hide Female Gens", true);
		RegisterBool(hideFemaleGens);
		var hideGensToggle = CreateToggle(hideFemaleGens, false);
		hideGensToggle.toggle.onValueChanged.AddListener((val) => {
			ReloadSkin();
		});
	
		CreateTextField(new JSONStorableString("drawMalePelvisText", "Draws the pelvis portion of male gen mesh. Surrounding vertices are morphed to align the edges."), false).height = 10;
		drawMalePelvis = new JSONStorableBool("Draw Male Pelvis", true);
		RegisterBool(drawMalePelvis);
		var drawMalePelvisToggle = CreateToggle(drawMalePelvis, false);
		drawMalePelvisToggle.toggle.onValueChanged.AddListener((val) => {
			ReloadSkin();
		});
		
		CreateTextField(new JSONStorableString("transparentShaderText", "Use a transparent shader for the male gens. Allows for an alpha mask and blending, but doesn't support tesselation."), false).height = 10;
		useTransparentShader = new JSONStorableBool("Use Transparent Shader", false);
		RegisterBool(useTransparentShader);
		var transparentShaderToggle = CreateToggle(useTransparentShader, false);
		transparentShaderToggle.toggle.onValueChanged.AddListener((val) => {
			ReloadSkin();
		});

		CreateTextField(new JSONStorableString("outerPelvisBase", "Which texture to use as the default for the female pelvis. Different ones work better for different skins."), false).height = 10;
		JSONStorableStringChooser outerPelvisSkinChooser = new JSONStorableStringChooser("outerPelvisBaseSkin", new List<string>{"Body", "Vagina"}, "Body", "Outer Pelvis Base Skin", (string baseSkin) => {
			useVaginaAsPelvisBase = baseSkin == "Vagina";
			overrideFemalePelvisTextures = true;
			ReloadSkin();
		});
		RegisterStringChooser(outerPelvisSkinChooser);
		
		CreateScrollablePopup(outerPelvisSkinChooser, false);
		
		genYOffetValue = AddSlider("Bone+Mesh Y Offset", genBoneYOffsetDefault, -0.1f, 0.1f, (float value) => {
			if(genYOffsetMorph != null){
				genYOffsetMorph.morphValue = value;
			}
		});
		
		genZOffetValue = AddSlider("Bone+Mesh Y Offset", genBoneZOffsetDefault, -0.1f, 0.1f, (float value) => {
			if(genZOffsetMorph != null){
				genZOffsetMorph.morphValue = value;
			}
		});
		
		genYMeshOffetValue = AddSlider("Mesh Y Offset", 0f, -0.1f, 0.1f, (float value) => {
			if(genYMeshOffsetMorph != null){
				genYMeshOffsetMorph.morphValue = value;
			}
		});
		
		genZMeshOffetValue = AddSlider("Mesh Z Offset", 0f, -0.1f, 0.1f, (float value) => {
			if(genZMeshOffsetMorph != null){
				genZMeshOffsetMorph.morphValue = value;
			}
		});
		
	
		
		HideMaterialOptionsIUI();
	}
	
	public JSONStorableFloat AddSlider(string name, float defaultVal, float min, float max, JSONStorableFloat.SetFloatCallback onChange = null, bool rangeAdjust = false, bool rightSide = false){
		JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
		storableFloat.storeType = JSONStorableParam.StoreType.Full;
		UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
		uiSlider.rangeAdjustEnabled = rangeAdjust;
		RegisterFloat(storableFloat);
		
		if(onChange != null){
			storableFloat.setCallbackFunction = onChange;
		}
		
		return storableFloat;
	}
	
	private IEnumerator InitMaterialUI(){
		// The material selction prefab doesn't do well with a layout,
		// so we're just going to remove it.
		while(scriptUI == null){
			if(this.UITransform != null){
				scriptUI = this.UITransform.GetComponentInChildren<MVRScriptUI>();
			}
			yield return null;
		}
		
		if(scriptUI == null){
			SuperController.LogMessage("Unable to load script UI.");
			yield break;
		}
		
		UnityEngine.UI.VerticalLayoutGroup layout = scriptUI.fullWidthUIContent.gameObject.GetComponent<UnityEngine.UI.VerticalLayoutGroup>();
		Destroy(layout);
		
		// Find a pre-existing material option UI to use as a template. 
		// Any would likely do...
		while( materialUIPrefab == null){
			foreach (GameObject go in Resources.FindObjectsOfTypeAll(typeof(GameObject)) as GameObject[]){
				if(go.name =="Mouth Materials") {
					materialUIPrefab = go;
					break;
				}
			}
			if(materialUIPrefab == null){
				yield return null;
			}
		}
		
		// Wait for the skin to load
		while(genMaterialOptions == null || pelvisMaterialOptions == null){
			yield return null;
		}
		
		try {	
			if(genMaterialOptionsUI == null){
				genMaterialOptionsUI = UnityEngine.Object.Instantiate<Transform>(materialUIPrefab.transform);
				genMaterialOptionsUI.SetParent(scriptUI.fullWidthUIContent,false);
			}
			genMaterialOptionsUI.gameObject.SetActive(false);
			genMaterialOptions.SetUI(genMaterialOptionsUI);
			genMaterialOptions.SyncAllParameters();
			genMaterialOptions.InitUI();
			
			
			if(pelvisMaterialOptionsUI == null){
				pelvisMaterialOptionsUI = UnityEngine.Object.Instantiate<Transform>(materialUIPrefab.transform);
				pelvisMaterialOptionsUI.SetParent(scriptUI.fullWidthUIContent,false);
			}
			pelvisMaterialOptionsUI.gameObject.SetActive(false);
			pelvisMaterialOptions.SetUI(pelvisMaterialOptionsUI);
			pelvisMaterialOptions.SyncAllParameters();
			pelvisMaterialOptions.InitUI();
			
			
			if(outerPelvisMaterialOptionsUI == null){
				outerPelvisMaterialOptionsUI = UnityEngine.Object.Instantiate<Transform>(materialUIPrefab.transform);
				outerPelvisMaterialOptionsUI.SetParent(scriptUI.fullWidthUIContent,false);
			}
			outerPelvisMaterialOptionsUI.gameObject.SetActive(false);
			outerPelvisMaterialOptions.SetUI(outerPelvisMaterialOptionsUI);
			outerPelvisMaterialOptions.SyncAllParameters();
			outerPelvisMaterialOptions.InitUI();
			
			// Back button.
			// Not the prettiest but it'll do
			Transform backButtonTransform = UnityEngine.Object.Instantiate<Transform>(manager.configurableButtonPrefab.transform);
			backButtonTransform.SetParent(scriptUI.fullWidthUIContent,false);
			
			UIDynamicButton backButton = backButtonTransform.GetComponent<UIDynamicButton>();
			backButton.label = "BACK";
			backButton.buttonColor = Color.green;
			backButton.button.onClick.AddListener(() => {
				HideMaterialOptionsIUI();
			});
			
			RectTransform rect = backButtonTransform.gameObject.GetComponent<RectTransform>();
			rect.anchorMax = new Vector2(0.15f,1f);
			
			scriptUI.fullWidthUIContent.gameObject.SetActive(false);
			
			RestoreSavedMaterialOptions();
			
			overrideFemalePelvisTextures = false;
		} catch (Exception e) {
			SuperController.LogError($"Error initializing material options ui. {e}");
		}
	}
	
	private void SetUpMorphs(){
		var bank = characterSelector.maleMorphBank2;
		ZeroDuplicateMorphs(bank);
		
		genYOffsetMorph = bank.GetMorphByUid($"{packagePath}Custom/Atom/Person/Morphs/male_genitalia/AltFuta Bone+Mesh Y-Offset.vmi");
		if(genYOffsetMorph == null){
			SuperController.LogError($"Unable to find morph AltFuta Bone+Mesh Y-Offset. Genitals may be positioned incorrectly.");
		} 
		
		genZOffsetMorph = bank.GetMorphByUid($"{packagePath}Custom/Atom/Person/Morphs/male_genitalia/AltFuta Bone+Mesh Z-Offset.vmi");
		if(genZOffsetMorph == null){
			SuperController.LogError($"Unable to find morph AltFuta Bone+Mesh Z-Offset. Genitals may be positioned incorrectly.");
		}
		
		genYMeshOffsetMorph = bank.GetMorphByUid($"{packagePath}Custom/Atom/Person/Morphs/male_genitalia/AltFuta Mesh Y-Offset.vmi");
		if(genYMeshOffsetMorph == null){
			SuperController.LogError($"Unable to find morph AltFuta Mesh Y-Offset. Genitals may be positioned incorrectly.");
		}
		
		genZMeshOffsetMorph = bank.GetMorphByUid($"{packagePath}Custom/Atom/Person/Morphs/male_genitalia/AltFuta Mesh Z-Offset.vmi");
		if(genZMeshOffsetMorph == null){
			SuperController.LogError($"Unable to find morph AltFuta Mesh Z-Offset. Genitals may be positioned incorrectly.");
		}

		genYOffetValue.setCallbackFunction(genYOffetValue.val);
		genZOffetValue.setCallbackFunction(genZOffetValue.val);
		genYMeshOffetValue.setCallbackFunction(genYMeshOffetValue.val);
		genZMeshOffetValue.setCallbackFunction(genZMeshOffetValue.val);
	}
	
	
	// Adapted from Everlaster's TittyMagic v70
	private void ZeroDuplicateMorphs(DAZMorphBank morphBank)
	{
		if(string.IsNullOrEmpty(packageId))
		{
			return;
		}
		var versionRegex = new Regex(@"\.\d+$", RegexOptions.Compiled);
		
		string packageName = versionRegex.Split(packageId)[0];
		foreach(var morph in morphBank.morphs)
		{
			if(
				morph.isInPackage &&
				morph.morphValue != 0 &&
				morph.packageUid.StartsWith(packageName) &&
				morph.packageUid != packageId
			)
			{
				morph.morphValue = 0;
			}
		}
	}

	private void SetUpMeshMorph(DAZMorph morph, Vector3 direction){	
		morph.region = "Morph/Genitalia";
		morph.min = -1f;
		morph.max = 1f;
		morph.visible = true;
		morph.isRuntime = true;
		morph.isDemandLoaded = false;
		
		// Vertex Deltas
		var deltas = new List<DAZMorphVertex>();
		for(int i = 0; i < genMesh.baseVertices.Length; i++){
			deltas.Add(new DAZMorphVertex(){vertex = i, delta = direction});
		}
		morph.deltas = deltas.ToArray();	
		morph.numDeltas = morph.deltas.Length;
		
		morph.formulas = new DAZMorphFormula[0];
	}
	
	private void SetUpBoneMorph(DAZMorph morph, DAZMorphFormulaTargetType targetType, Vector3 direction){
		SetUpMeshMorph(morph,direction);

		// Bone Morphs
		var boneFormulas = new List<DAZMorphFormula>();
		boneFormulas.Add(new DAZMorphFormula(){target = "Gen1", multiplier = 1f, targetType = targetType});
		boneFormulas.Add(new DAZMorphFormula(){target = "Gen2", multiplier = 1f, targetType = targetType});
		boneFormulas.Add(new DAZMorphFormula(){target = "Gen3", multiplier = 1f, targetType = targetType});
		boneFormulas.Add(new DAZMorphFormula(){target = "Testes", multiplier = 1f, targetType = targetType});
		
		morph.formulas = boneFormulas.ToArray();
	}
	
	private DAZMorphFormula MorphBoneFormula(string boneName, DAZMorphFormulaTargetType targetType){
		var formula = new DAZMorphFormula();
		formula.targetType = targetType;
		formula.target = boneName;
		formula.multiplier = 1f;
		return formula;
	}
	
	private MaterialOptionsWrapper CreateMaterialOptions(int matIndex, Material defaultMaterial){
		var matOpts = newCharacterContainer.AddComponent<MaterialOptionsWrapper>();
		
		matOpts.textureGroup1 = new MaterialOptionTextureGroup();
		matOpts.SetCustomTextureFolder("Custom/Atom/Person/Textures");

		matOpts.materialForDefaults = defaultMaterial;

		matOpts.paramMaterialSlots = new int[]{matIndex};
		matOpts.textureGroup1.materialSlots = matOpts.paramMaterialSlots;
		
		matOpts.skin = skin;
		
		matOpts.enabled = true;
		matOpts.SyncAllParameters();
		return matOpts;
	}
	
	private void ShowMaterialOptionsUI(Transform matOptsUI){
		if(matOptsUI == null)
			return;
		
		if(genMaterialOptionsUI != null && genMaterialOptionsUI != matOptsUI)
			genMaterialOptionsUI.gameObject.SetActive(false);
		
		if(pelvisMaterialOptionsUI != null && pelvisMaterialOptionsUI != matOptsUI)
			pelvisMaterialOptionsUI.gameObject.SetActive(false);
		
		if(outerPelvisMaterialOptionsUI != null && outerPelvisMaterialOptionsUI != matOptsUI)
			outerPelvisMaterialOptionsUI.gameObject.SetActive(false);
		
		
		scriptUI.leftUIContent.gameObject.SetActive(false);
		scriptUI.rightUIContent.gameObject.SetActive(false);
		scriptUI.fullWidthUIContent.gameObject.SetActive(true);
		
		matOptsUI.gameObject.SetActive(true);
	}
	
	private void HideMaterialOptionsIUI(){
		if(genMaterialOptionsUI != null)
			genMaterialOptionsUI.gameObject.SetActive(false);
		
		if(pelvisMaterialOptionsUI != null)
			pelvisMaterialOptionsUI.gameObject.SetActive(false);
		
		if(outerPelvisMaterialOptionsUI != null)
			outerPelvisMaterialOptionsUI.gameObject.SetActive(false);
			
		
		if(scriptUI != null){
			scriptUI.leftUIContent.gameObject.SetActive(true);
			scriptUI.rightUIContent.gameObject.SetActive(true);
			scriptUI.fullWidthUIContent.gameObject.SetActive(false);
		}
	}
	
	private void RestoreSavedMaterialOptions(){
		try {

			
			if(oldSkinOpts != null && newCharacter != null){
				foreach(var json in oldSkinOpts){
					foreach(var options in newCharacter.GetComponentsInChildren<DAZCharacterMaterialOptions>()){
						string id = json["id"];
						if(id != null && id == options.storeId){
							options.RestoreFromJSON(json);
						}
					}
				}
			}
			
			var textureControl = newCharacter.GetComponentInChildren<DAZCharacterTextureControl>();
			if(textureControl != null && oldTextureControlJson != null){
				textureControl.RestoreFromJSON(oldTextureControlJson);
			}
			
			var skinMaterialOpts = containingAtom.GetStorableByID("skin") as DAZCharacterMaterialOptions;
			genMaterialOptions.CopyMaterialParams(skinMaterialOpts);
			pelvisMaterialOptions.CopyMaterialParams(skinMaterialOpts);
			outerPelvisMaterialOptions.CopyMaterialParams(skinMaterialOpts);
			
			if(!overrideFemalePelvisTextures){
				SyncFemalePelvisTextures(outerPelvisMaterialOptions, textureControl);
			}
			
			// We don't want the textures in character skin options to apply to the pelvis
			// which they do by default.
			pelvisMaterialOptions.ClearCustomTextures();
			
			
			if(savedSceneConfig != null){
				var genMatOptionsJson = savedSceneConfig["genMaterialOpts"] as JSONClass;
				if(genMatOptionsJson != null && genMaterialOptions != null){
					genMaterialOptions.RestoreFromJSON(genMatOptionsJson);
				}
				
				var pelvisMatOptionsJson = savedSceneConfig["pelvisMaterialOpts"] as JSONClass;
				if(pelvisMatOptionsJson != null && pelvisMaterialOptions != null){
					pelvisMaterialOptions.RestoreFromJSON(pelvisMatOptionsJson);
				}
				
				var outerPelvisMatOptionsJson = savedSceneConfig["outerPelvisMaterialOpts"] as JSONClass;
				if(outerPelvisMatOptionsJson != null && outerPelvisMaterialOptions != null){
					outerPelvisMaterialOptions.RestoreFromJSON(outerPelvisMatOptionsJson);
				}
			}
			
			genMaterialOptions.SetParams();
			pelvisMaterialOptions.SetParams();
			outerPelvisMaterialOptions.SetParams();			
			
			if(overrideFemalePelvisTextures){
				SyncFemalePelvisTextures(outerPelvisMaterialOptions, textureControl);
			}
		} catch (Exception e) {
			SuperController.LogError($"Error copying skin options from previous skin. {e}");
		}
	}
	
	
	/*
		The texture control that controls the body's textures from the "Skin Textures" tab
		doesn't go through material options. It sets textures on the material directly. 
		
	*/
	private void SyncFemalePelvisTextures(MaterialOptions pelvisOpts, DAZCharacterTextureControl textureControl){
		var bodypart = useVaginaAsPelvisBase ? "genitals" : "torso";
		pelvisOpts.GetUrlJSONParam("customTexture" + pelvisOpts.textureGroup1.textureName).val =  textureControl.GetUrlJSONParam(bodypart + "DiffuseUrl").val;
		pelvisOpts.GetUrlJSONParam("customTexture" + pelvisOpts.textureGroup1.secondaryTextureName).val =  textureControl.GetUrlJSONParam(bodypart + "SpecularUrl").val;
		pelvisOpts.GetUrlJSONParam("customTexture" + pelvisOpts.textureGroup1.thirdTextureName).val =  textureControl.GetUrlJSONParam(bodypart + "GlossUrl").val;
		//Fourth is alpha
		pelvisOpts.GetUrlJSONParam("customTexture" + pelvisOpts.textureGroup1.fifthTextureName).val =  textureControl.GetUrlJSONParam(bodypart + "NormalUrl").val;
		pelvisOpts.GetUrlJSONParam("customTexture" + pelvisOpts.textureGroup1.sixthTextureName).val =  textureControl.GetUrlJSONParam(bodypart + "DecalUrl").val;
		
	}
	
	private string GetPathToObject(GameObject go){
		 if (go.transform.parent == null || go.transform.parent == containingAtom.gameObject.transform)
			 return "/" + go.name;
		 return GetPathToObject(go.transform.parent.gameObject) + "/" + go.name;
	}

	/* Change the status of the male portions of the model.
		Specifically we only want the rigid bodies and controllers
	*/
	private void SetMaleTransformsActive(bool active){
		foreach(GameObject go in maleObjects){
			go.SetActive(active);
			SyncAutoColliderIgnore(go);
		}
		
		foreach(var ac in maleGen1.GetComponentsInChildren<AutoCollider>(true)){
			ac.gameObject.SetActive(active);
		}
	}
	
	private void UpdteAutoColliders(){
		foreach( var entry in maleAutoColliderConfigs){
			var ac = entry.Key;
			var resizeTrigger = ac.resizeTrigger;
			ac.resizeTrigger = AutoCollider.ResizeTrigger.Always;
			ac.AutoColliderSizeSet(true);
			ac.resizeTrigger = resizeTrigger;
		}
	}
	
	private void SyncAutoColliderIgnore(GameObject go){
		if(go.activeInHierarchy){
			foreach(AutoCollider ac in genAutoColliderIgnoreList){
				var ignoreList = ac.ignoreColliders.ToList();
				if(!ignoreList.Contains(go.transform)){
					ignoreList.Add(go.transform);
					ac.ignoreColliders = ignoreList.ToArray();
					ac.InitColliders();
				}
			}
		}
	}
	
	private void Update(){
		
		// For whatever reason, putting on clothes resets this;
		if( skin != null){
			skin.materialsEnabled[29] = drawMalePelvis.val;
			skin.materialsEnabled[28]= !hideFemaleGens.val;
			skin.materialsEnabled[31] = hideFemaleGens.val;
		}
		
		if(characterRun.morphBank3 != null){
			// Fix these settings in case they got changed somehow
			if(characterRun.morphBank3 != characterSelector.maleMorphBank2){
				characterRun.morphBank3 = characterSelector.maleMorphBank2;
			}

			if(characterRun.morphBank3.connectedMesh != null){
				characterRun.morphBank3.connectedMesh = null;
			}
			
			// For whatever reason, the threaded morph logic in DAZCharacterRun
			// just doesn't bother with the 2nd graft. 
			// So we'll just do it manually here
			if ( characterRun.morphBank3.bonesDirty)	{
				characterRun.bones.SetMorphedTransform(false);
				characterRun.morphBank3.bonesDirty = false;
			}
			if (characterRun.morphBank3.boneRotationsDirty != null)
			{
				foreach (var entry in characterRun.morphBank3.boneRotationsDirty){
					entry.Key.SyncMorphBoneRotations(false);
				}
				characterRun.morphBank3.boneRotationsDirty.Clear();
			}
		}
		
		if (Input.GetMouseButtonDown(1)){
			SetMaleTransformsActive(true);
		}
	}
	
	
	
	private void UnloadSkin(){
		SetMaleTransformsActive(false);
		
		skin = null;
		genMaterialOptions = null;
		pelvisMaterialOptions = null;
		outerPelvisMaterialOptions = null;
		
		if(maleCharacterContainer != null){
			Destroy(maleCharacterContainer);
		}
		if(newCharacterContainer != null){
			Destroy(newCharacterContainer);
		}
		
		maleCharacterContainer = null;
		newCharacterContainer = null;
	}
	
	private void OnEnable(){
		ReloadSkin();
	}
	
	private void OnDisable(){
		if(characterSelector != null && previousSelected != null && characterSelector.selectedCharacter == newCharacter){
			characterSelector.selectedCharacter = previousSelected;
		}
	}
	
	private void OnDestroy()
	{	
		SetMaleTransformsActive(false);
		UnloadSkin();
	}
		
	List<GameObject> GetAllChildren(GameObject parent, List<GameObject> objectList = null)
	{
		if (objectList == null) objectList = new List<GameObject>();
	  
		foreach (Transform child in parent.transform) {
			objectList.Add(child.gameObject);
			GetAllChildren(child.gameObject, objectList);
		}
		return objectList;
	}
	
	public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true) {
		savedSceneConfig = jc;
		
		base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
		
		if(jc["baseSkinMaterialOpts"] != null){
			var optsList = new List<JSONClass>();
			foreach(var optsJson in jc["baseSkinMaterialOpts"].AsArray){
				optsList.Add(optsJson as JSONClass);
			}
			oldSkinOpts = optsList.ToArray();
		}
		
		if(jc["baseTextureControl"] != null){
			oldTextureControlJson = jc["baseTextureControl"].AsObject;
		}
	}
	
	public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
	{
		var jsonClass = base.GetJSON(true, true, true);
		
		jsonClass.Add("genMaterialOpts", genMaterialOptions.GetJSON(true,true,true)); 
		
		jsonClass.Add("pelvisMaterialOpts", pelvisMaterialOptions.GetJSON(true,true,true)); 
		
		jsonClass.Add("outerPelvisMaterialOpts", outerPelvisMaterialOptions.GetJSON(true,true,true)); 
		
		JSONArray baseSkinMaterialOpts = new JSONArray();
		foreach(var options in newCharacter.GetComponentsInChildren<DAZCharacterMaterialOptions>()){
			baseSkinMaterialOpts.Add(options.GetJSON(true,true,true));
		}		
		jsonClass.Add("baseSkinMaterialOpts", baseSkinMaterialOpts); 
		
		
		var textureControl = newCharacter.GetComponentInChildren<DAZCharacterTextureControl>();
		jsonClass.Add("baseTextureControl", textureControl.GetJSON(true,true,true)); 
		
		return jsonClass;
	}
	
	// Function taken from VAMDeluxe's code
	public static string GetPluginPath(MVRScript plugin, Atom atom){
		SuperController.singleton.currentSaveDir = SuperController.singleton.currentLoadDir;
		string pluginId = plugin.storeId.Split('_')[0];
		MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
		string pathToScriptFile = manager.GetJSON(true, true)["plugins"][pluginId].Value;
		string pathToScriptFolder = pathToScriptFile.Substring(0, pathToScriptFile.LastIndexOfAny(new char[] { '/', '\\' }));
		return pathToScriptFolder;
	}
	
	
	//MacGruber / Discord 20.10.2020
	//Get path prefix of the package that contains this plugin
	private static string GetPackageId(MVRScript script)
	{
		string id = script.name.Substring(0, script.name.IndexOf('_'));
		string filename = script.manager.GetJSON()["plugins"][id].Value;
		int idx = filename.IndexOf(":/", StringComparison.Ordinal);
		return idx >= 0 ? filename.Substring(0, idx) : null;
	}
}
}